"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.BaseType = exports.Permissions = void 0;

var _suryaGqlScalar = require("surya-gql-scalar");

var _util = require("util");

const Permissions = {
  Public: "Public",
  // can be accessed by any user
  Secure: "Secure",
  // can be accessed by a user with Secure privilege on object type
  Personal: "Personal",
  // can be accessed by a user who owns the immediate object
  Internal: "Internal" // is not exposed on external api, but is on the API returned object

};
exports.Permissions = Permissions;

class BaseType {
  constructor(data) {
    if (data) {
      this.loadData(data);
    }
  }
  /**
   * Override typeSpec() to return definition of type global properties and fields.
   */


  typeSpec() {
    return {
      name: "BaseType",
      extends: null,
      instantiable: false,
      timestamps: false,
      permissions: {
        Secure: ["auth"]
      },
      fields: {}
    };
  }

  getOwnerField() {
    const spec = this.typeSpec();
    let ownerFieldName = null;

    if (spec && spec.fields && (0, _util.isObject)(spec.fields)) {
      for (let fieldName of Object.keys(spec.fields)) {
        const fieldSpec = spec.fields[fieldName];

        if (fieldSpec.ownerId) {
          ownerFieldName = fieldName;
          break;
        }
      }
    }

    return ownerFieldName;
  }

  getKeyField() {
    const spec = this.typeSpec();
    let keyFieldName = null;

    if (spec && spec.fields && (0, _util.isObject)(spec.fields)) {
      for (let fieldName of Object.keys(spec.fields)) {
        const fieldSpec = spec.fields[fieldName];

        if (fieldSpec.key) {
          keyFieldName = fieldName;
          break;
        }
      }
    }

    return keyFieldName;
  }

  getOwnerId() {
    const ownerFieldName = this.getOwnerField();
    return this[ownerFieldName];
  }
  /**
   * Default implementation of isOwner logic.
   * Override this logic in cases where the userId of the owner is not directly stored as a data field of the object
   *
   * @param {*} userId
   */


  isOwner(userId) {
    return this.getOwnerId() == userId;
  }

  getPermissions(permType) {
    const typeSpec = this.typeSpec();

    if (typeSpec.permissions) {
      return typeSpec.permissions[permType];
    }

    return [];
  }
  /**
   * Apply permission filters
   * Expect to be called only if user is not Secure (has permissions to Secure fields.)
   * filter out secure fields
   * filter out Personal fields unless the user is the owner
   * @param {*} authUserId
   */


  applyReadPermissions(authUserId) {
    const isOwner = this.isOwner(authUserId);
    let filtered = Object.create(this);
    filtered.loadData(this);
    const className = this.constructor.name;
    const spec = this.typeSpec();

    if (spec && spec.fields && (0, _util.isObject)(spec.fields)) {
      Object.keys(spec.fields).forEach(fieldName => {
        const fieldSpec = spec.fields[fieldName];

        this._filterDataField(isOwner, fieldName, fieldSpec, filtered);
      });
    }

    return filtered;
  }

  _filterDataField(isOwner, fieldName, fieldSpec, filtered) {
    const readPermission = fieldSpec.readPermission || Permissions.Public;
    const filter = readPermission == Permissions.Secure || readPermission == Permissions.Personal && !isOwner;
    if (filter) filtered[fieldName] = null;
  }
  /**
   * Load basic JSON data per spec
   * @param {*} data
   */


  loadData(data) {
    if (data) {
      const spec = this.typeSpec();

      if (spec && spec.fields && (0, _util.isObject)(spec.fields)) {
        Object.keys(spec.fields).forEach(fieldName => {
          const fieldSpec = spec.fields[fieldName];

          this._loadDataField(fieldName, fieldSpec, data);
        });
      }
    }

    return this;
  }

  _loadDataField(fieldName, fieldSpec, data) {
    const useFieldName = fieldSpec.type == _suryaGqlScalar.ScalarTypes.ObjectRef ? fieldSpec.refField : fieldName;
    this[useFieldName] = data[useFieldName];
  }
  /**
   * Load Data as per asName.
   * This would be useful for loading up data for return that has perumted names
   * @param {*} data
   */


  loadNative(data) {
    if (data) {
      const spec = this.typeSpec();

      if (spec && spec.fields && (0, _util.isObject)(spec.fields)) {
        Object.keys(spec.fields).forEach(fieldName => {
          const fieldSpec = spec.fields[fieldName];

          this._loadFromNativeField(data, this, fieldName, fieldSpec, data);
        });
      }
    }

    return this;
  }

  _loadFromNativeField(data, obj, fieldName, fieldSpec) {
    const fromFieldName = fieldSpec.asName ? fieldSpec.asName : fieldName;
    let fieldVal = data[fromFieldName];

    if (fieldSpec.type == _suryaGqlScalar.ScalarTypes.Object && fieldVal) {
      const subTypeName = fieldSpec.objectType;
      const subType = this.childTypes()[subTypeName];

      if (subType) {
        if (fieldSpec.array) {
          fieldVal = fieldVal.map(item => {
            const subObj = new subType();
            return subObj.loadNative(item);
          });
        } else {
          const subObj = new subType();
          fieldVal = subObj.loadNative(fieldVal);
        }
      } else {
        throw new Error(`type not found: ${subTypeName}`);
      }
    }

    obj[fieldName] = fieldVal;
  }
  /**
   * Return an object (JSON) that has the data files permuted as per the asName parameters.
   *
   */


  toNativeJson() {
    let nativeObj = {};
    const spec = this.typeSpec();

    if (spec && spec.fields && (0, _util.isObject)(spec.fields)) {
      Object.keys(spec.fields).forEach(fieldName => {
        const fieldSpec = spec.fields[fieldName];

        this._loadToNativeField(nativeObj, this, fieldName, fieldSpec);
      });
    }

    return nativeObj;
  }

  _loadToNativeField(obj, data, fieldName, fieldSpec) {
    const toFieldName = fieldSpec.asName ? fieldSpec.asName : fieldName;
    let fieldVal = data[fieldName];

    if (fieldSpec.type == _suryaGqlScalar.ScalarTypes.Object && fieldVal) {
      const subTypeName = fieldSpec.objectType;
      const subType = this.childTypes()[subTypeName];

      if (subType) {
        if (fieldSpec.array) {
          fieldVal = fieldVal.map(item => {
            const subObj = new subType(item);
            return subObj.toNativeJson();
          });
        } else {
          const subObj = new subType(fieldVal);
          fieldVal = subObj.toNativeJson();
        }
      } else {
        throw new Error(`type not found: ${subTypeName}`);
      }
    }

    obj[toFieldName] = fieldVal;
  }
  /**
   * Load data from mongo document with necessary modifications
   * TODO: Move this to and adapter model hosted in the API
   * @param {*} data
   */


  loadDocument(data) {
    const spec = this.typeSpec();

    if (spec && spec.fields && (0, _util.isObject)(spec.fields)) {
      Object.keys(spec.fields).forEach(fieldName => {
        const fieldSpec = spec.fields[fieldName];

        this._loadDocField(fieldName, fieldSpec, data);
      });
    }

    return this;
  }

  _loadDocField(fieldName, fieldSpec, doc) {
    if (!doc) return;
    const data = doc._doc; // Otherwise non-string types are not getting.

    let useFieldName = fieldName;

    if (!!fieldSpec.key) {
      this[fieldName] = data._id.toString();
    } else {
      if (fieldSpec.type == _suryaGqlScalar.ScalarTypes.ObjectRef) {
        useFieldName = fieldSpec.refField;
      }

      let value = data[useFieldName];

      switch (fieldSpec.type) {
        case _suryaGqlScalar.ScalarTypes.DateTime:
          // do nothing. we expect a Date object in the Date type
          break;
      }

      this[useFieldName] = value;
    }
  }

  toGraphQLSchema() {
    const spec = this.typeSpec();
    this.schemaTypeName = spec.name;

    const typeSchema = this._genType(spec);

    const inputSchema = this._genInput(spec);

    const setResponseSchema = this._genSetResponseType(spec);

    const mutationResponseSchema = this._genMutationResponseType(spec);

    const specInputSchemas = this._genSpecInputs(spec);

    const inlineScalars = this._genScalars(spec); // console.log(`Scalars [${this.schemaTypeName}]: \n> ${inlineScalars};`)


    const schemas = typeSchema + "\n" + inputSchema + "\n" + setResponseSchema + "\n" + mutationResponseSchema + "\n" + specInputSchemas + "\n" + inlineScalars + "\n";
    return schemas;
  }

  _genType(spec) {
    return `type ${spec.name} {
${this._genFields(spec.fields, false, null)}
}`;
  }

  _genInput(spec) {
    return `input ${spec.name}Input {
${this._genFields(spec.fields, true, null)}
}`;
  }

  _genSpecInputs(spec) {
    const inputNames = spec.inputs || [];
    let specInputs = [];

    for (let n = 0; n < inputNames.length; n++) {
      const inputName = inputNames[n];
      const specInput = `input ${inputName} {
        ${this._genFields(spec.fields, true, inputName)}
        }`;
      specInputs.push(specInput);
    }

    return specInputs.join("/n");
  }

  _genSetResponseType(spec) {
    if (!!spec.instantiable) {
      return `type ${spec.name}SetResponse {
  items: [${spec.name}]!
  total: Int
  count: Int
  offset: Int
}`;
    } else {
      return "";
    }
  }

  _genMutationResponseType(spec) {
    if (!!spec.instantiable) {
      return `type ${spec.name}MutationResponse {
  success: Boolean!
  message: String
  data: ${spec.name}
}
`;
    } else {
      return "";
    }
  }

  _genFields(fields, input = false, specInput = null) {
    let fstr = Object.getOwnPropertyNames(fields).map(name => {
      const typedef = fields[name];

      const {
        fieldName,
        fieldType,
        include,
        params,
        scalar
      } = this._typeMap(name, typedef, input, specInput);

      if (include) {
        return `    ${fieldName}${params}: ${fieldType}\n`;
      } else {
        return "";
      }
    }).join(" ");
    return fstr;
  }

  _genScalars(spec) {
    const fields = spec.fields;
    let fstr = Object.getOwnPropertyNames(fields).map(name => {
      const typedef = fields[name];

      const {
        fieldName,
        fieldType,
        include,
        params,
        scalar
      } = this._typeMap(name, typedef, false, "");

      if (scalar.length > 0) {
        return `${scalar}\n`;
      }
    }).join(" ");
    return fstr;
  }

  _typeMap(fieldName, typeDef, input = false, specInput = null) {
    const typeName = typeDef.type;
    const required = !!typeDef.required;
    const permission = typeDef.readPermission || Permissions.Public;
    let include = (!input || input && typeDef.input !== false) && permission != Permissions.Internal;
    let pre = "";
    let post = "";
    let params = "";

    if (typeDef.array) {
      pre = "[";
      post = "]";
    }

    const bang = required ? "!" : "";
    let useFieldName = fieldName;
    let scalar = ""; // auxiliar scalars such as enums

    let gqlType = "String";

    switch (typeName) {
      case _suryaGqlScalar.ScalarTypes.String:
        gqlType = pre + "String" + bang + post;
        break;

      case _suryaGqlScalar.ScalarTypes.DateTime:
        gqlType = pre + "Date" + bang + post;
        break;

      case _suryaGqlScalar.ScalarTypes.Integer:
        gqlType = pre + "Int" + bang + post;
        break;

      case _suryaGqlScalar.ScalarTypes.Float:
        gqlType = pre + "Float" + bang + post;
        break;

      case _suryaGqlScalar.ScalarTypes.ID:
        gqlType = pre + "ID" + bang + post;
        break;

      case _suryaGqlScalar.ScalarTypes.Boolean:
        gqlType = pre + "Boolean" + bang + post;
        break;

      case _suryaGqlScalar.ScalarTypes.Object:
        let typeName = typeDef.objectType + (input ? "Input" : "");
        gqlType = pre + typeName + bang + post;
        break;

      case _suryaGqlScalar.ScalarTypes.ObjectRef:
        const paramType = typeDef.params || "";

        if (input) {
          const specInputs = typeDef.inputs || [];

          if (specInput && specInputs.includes(specInput)) {
            // embed as if it were an input object
            let typeName = typeDef.objectType + (input ? "Input" : "");
            gqlType = pre + typeName + bang + post;
          } else {
            // A usual input id reference
            if (typeDef.refField) {
              useFieldName = typeDef.refField;
              gqlType = pre + typeDef.refFieldType + bang + post;
            } else {
              // if no refField specified, we assume it's not insertable
              include = false;
            }
          }
        } else {
          gqlType = pre + typeDef.objectType + bang + post;

          if (paramType == "query") {
            params = "(offset: Int, limit: Int, filter: String, sort: String)";
          }
        }

        break;

      case _suryaGqlScalar.ScalarTypes.Enum:
        const enumName = `${this.schemaTypeName}_${useFieldName}_ENUM`.toUpperCase();
        const scalarEnum = typeDef.enum.join("\n");
        scalar = `enum ${enumName} {
${scalarEnum}
}`;
        gqlType = pre + enumName + bang + post;
        break;
    }

    return {
      fieldName: useFieldName,
      fieldType: gqlType,
      include: include,
      params,
      scalar
    };
  }

}

exports.BaseType = BaseType;
//# sourceMappingURL=baseType.js.map